import torch
from torch.autograd import Function

try:
    import mhc_cuda
except ImportError:
    raise ImportError(
        "mhc_cuda not found. Please install the CUDA extension by running:\n"
        "pip install -e ."
    )


class SinkhornKnoppFunction(Function):
    @staticmethod
    def forward(ctx, inp, num_iters, eps):
        out = mhc_cuda.sinkhorn_knopp_fwd(inp.contiguous(), num_iters, eps)
        ctx.save_for_backward(out, inp)
        ctx.num_iters = num_iters
        ctx.eps = eps
        return out

    @staticmethod
    def backward(ctx, grad_output):
        out, inp = ctx.saved_tensors
        d_inp = mhc_cuda.sinkhorn_knopp_bwd(
            grad_output.contiguous(), out, inp, ctx.num_iters, ctx.eps
        )
        return d_inp, None, None


class RMSNormFunction(Function):
    @staticmethod
    def forward(ctx, inp, weight, eps):
        out, rms = mhc_cuda.rmsnorm_fwd(inp.contiguous(), weight.contiguous(), eps)
        ctx.save_for_backward(inp, weight, rms)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        inp, weight, rms = ctx.saved_tensors
        d_inp, d_weight = mhc_cuda.rmsnorm_bwd(
            grad_output.contiguous(), inp, weight, rms
        )
        return d_inp, d_weight, None


def sinkhorn_knopp(inp, num_iters=20, eps=1e-8):
    return SinkhornKnoppFunction.apply(inp.float(), num_iters, eps)


def rmsnorm(inp, weight, eps=1e-5):
    return RMSNormFunction.apply(inp.bfloat16(), weight.bfloat16(), eps)


class MHCLayerFunction(Function):
    @staticmethod
    def forward(
        ctx, x_expanded, rmsnorm_weight, H_pre, H_post, H_res, sinkhorn_iters, eps
    ):
        (
            output,
            rms,
            x_agg_bf16,
            H_pre_activated,
            H_post_activated,
            M,
            y_norm_bf16,
        ) = mhc_cuda.mhc_layer_fwd(
            x_expanded.contiguous(),
            rmsnorm_weight.contiguous(),
            H_pre.contiguous(),
            H_post.contiguous(),
            H_res.contiguous(),
            sinkhorn_iters,
            eps,
        )

        ctx.save_for_backward(
            x_expanded,
            rmsnorm_weight,
            rms,
            x_agg_bf16,
            H_pre_activated,
            H_post_activated,
            M,
            y_norm_bf16,
            H_res,
        )
        ctx.sinkhorn_iters = sinkhorn_iters
        ctx.eps = eps
        return output

    @staticmethod
    def backward(ctx, grad_output):
        (
            x_expanded,
            rmsnorm_weight,
            rms,
            x_agg_bf16,
            H_pre_activated,
            H_post_activated,
            M,
            y_norm_bf16,
            H_res,
        ) = ctx.saved_tensors

        d_x, d_rmsnorm_weight, d_H_pre, d_H_post, d_H_res = mhc_cuda.mhc_layer_bwd(
            grad_output.contiguous(),
            x_expanded.contiguous(),
            rmsnorm_weight.contiguous(),
            rms.contiguous(),
            x_agg_bf16.contiguous(),
            H_pre_activated.contiguous(),
            H_post_activated.contiguous(),
            M.contiguous(),
            y_norm_bf16.contiguous(),
            H_res.contiguous(),
            ctx.sinkhorn_iters,
            ctx.eps,
        )

        return d_x, d_rmsnorm_weight, d_H_pre, d_H_post, d_H_res, None, None


def mhc_layer_fused(
    x_expanded, rmsnorm_weight, H_pre, H_post, H_res, sinkhorn_iters=20, eps=1e-5
):
    return MHCLayerFunction.apply(
        x_expanded.float(),
        rmsnorm_weight.bfloat16(),
        H_pre.float(),
        H_post.float(),
        H_res.float(),
        sinkhorn_iters,
        eps,
    )


def mhc_layer_fused_inference(
    x_expanded, rmsnorm_weight, H_pre, H_post, H_res, sinkhorn_iters=20, eps=1e-5
):
    # inference focused forward pass -> no backward support for maximum speed
    return mhc_cuda.mhc_layer_fwd_inference(
        x_expanded.float().contiguous(),
        rmsnorm_weight.bfloat16().contiguous(),
        H_pre.float().contiguous(),
        H_post.float().contiguous(),
        H_res.float().contiguous(),
        sinkhorn_iters,
        eps,
    )


class MHCLayerDynamicFunction(Function):
    @staticmethod
    def forward(
        ctx,
        x_expanded,
        rmsnorm_weight,
        phi_pre,
        phi_post,
        phi_res,
        alpha_pre,
        alpha_post,
        alpha_res,
        b_pre,
        b_post,
        b_res,
        sinkhorn_iters,
        eps,
    ):
        n = phi_pre.size(0)
        phi_res_view = phi_res.view(n * n, -1)
        nC = phi_pre.size(1)
        out_dim = n + n + n * n
        phi_concat = torch.empty(
            (out_dim, nC),
            device=phi_pre.device,
            dtype=torch.bfloat16,
        )
        # Avoid a large float32 concat temporary to reduce peak memory.
        phi_concat[:n].copy_(phi_pre)
        phi_concat[n : 2 * n].copy_(phi_post)
        phi_concat[2 * n :].copy_(phi_res_view)
        alpha_pre_val = (
            float(alpha_pre.item()) if torch.is_tensor(alpha_pre) else float(alpha_pre)
        )
        alpha_post_val = (
            float(alpha_post.item())
            if torch.is_tensor(alpha_post)
            else float(alpha_post)
        )
        alpha_res_val = (
            float(alpha_res.item()) if torch.is_tensor(alpha_res) else float(alpha_res)
        )

        (
            output,
            rms,
            x_agg_bf16,
            H_pre_activated,
            H_post_activated,
            M,
            y_norm_bf16,
            x_flat_bf16,
            rms_h,
        ) = mhc_cuda.mhc_layer_fwd_dynamic(
            x_expanded.contiguous(),
            rmsnorm_weight.contiguous(),
            phi_concat,
            alpha_pre_val,
            alpha_post_val,
            alpha_res_val,
            b_pre.contiguous(),
            b_post.contiguous(),
            b_res.contiguous(),
            sinkhorn_iters,
            eps,
        )

        ctx.save_for_backward(
            x_expanded,
            rmsnorm_weight,
            rms,
            x_agg_bf16,
            H_pre_activated,
            H_post_activated,
            M,
            y_norm_bf16,
            x_flat_bf16,
            rms_h,
            phi_pre,
            phi_post,
            phi_res,
            b_pre,
            b_post,
            b_res,
            alpha_pre,
            alpha_post,
            alpha_res,
        )
        ctx.sinkhorn_iters = sinkhorn_iters
        ctx.eps = eps
        return output

    @staticmethod
    def backward(ctx, grad_output):
        (
            x_expanded,
            rmsnorm_weight,
            rms,
            x_agg_bf16,
            H_pre_activated,
            H_post_activated,
            M,
            y_norm_bf16,
            x_flat_bf16,
            rms_h,
            phi_pre,
            phi_post,
            phi_res,
            b_pre,
            b_post,
            b_res,
            alpha_pre,
            alpha_post,
            alpha_res,
        ) = ctx.saved_tensors

        d_x, d_rmsnorm_weight, d_H_pre, d_H_post, d_M = mhc_cuda.mhc_layer_bwd_dynamic(
            grad_output.contiguous(),
            x_expanded.contiguous(),
            rmsnorm_weight.contiguous(),
            rms.contiguous(),
            x_agg_bf16.contiguous(),
            H_pre_activated.contiguous(),
            H_post_activated.contiguous(),
            M.contiguous(),
            y_norm_bf16.contiguous(),
            ctx.eps,
        )

        with torch.no_grad():
            B, n, C = x_expanded.shape
            nC = n * C

            x_flat = x_flat_bf16.float()
            rms_h_f32 = rms_h.float()
            rms_inv = 1.0 / rms_h_f32
            rms_inv2 = rms_inv * rms_inv

            phi_pre_f32 = phi_pre.float()
            phi_post_f32 = phi_post.float()
            phi_res_f32 = phi_res.float()

            p_pre = x_flat @ phi_pre_f32.t()
            p_post = x_flat @ phi_post_f32.t()
            p_res_flat = x_flat @ phi_res_f32.t()
            p_res = p_res_flat.view(B, n, n)

            H_pre_act = H_pre_activated.float()
            H_post_act = H_post_activated.float()

            d_tilde_pre = d_H_pre * H_pre_act * (1.0 - H_pre_act)
            d_tilde_post = d_H_post * H_post_act * (1.0 - H_post_act / 2.0)

            alpha_pre_f32 = alpha_pre.float()
            alpha_post_f32 = alpha_post.float()
            alpha_res_f32 = alpha_res.float()
            b_res_f32 = b_res.float()

            tilde_res = alpha_res_f32 * p_res * rms_inv.view(B, 1, 1) + b_res_f32
            tilde_res = torch.clamp(tilde_res, max=20.0)
            H_res_exp = torch.exp(tilde_res)

            d_H_res_exp = mhc_cuda.sinkhorn_knopp_bwd_batched(
                d_M.contiguous(),
                H_res_exp.contiguous(),
                ctx.sinkhorn_iters,
                ctx.eps,
            )
            d_tilde_res = d_H_res_exp * H_res_exp

            d_b_pre = d_tilde_pre.sum(dim=0)
            d_b_post = d_tilde_post.sum(dim=0)
            d_b_res = d_tilde_res.sum(dim=0)

            d_alpha_pre = (d_tilde_pre * (p_pre * rms_inv[:, None])).sum()
            d_alpha_post = (d_tilde_post * (p_post * rms_inv[:, None])).sum()
            d_alpha_res = (d_tilde_res * (p_res * rms_inv.view(B, 1, 1))).sum()

            d_p_pre = d_tilde_pre * (alpha_pre_f32 * rms_inv[:, None])
            d_p_post = d_tilde_post * (alpha_post_f32 * rms_inv[:, None])
            d_p_res = d_tilde_res * (alpha_res_f32 * rms_inv.view(B, 1, 1))
            d_p_res_flat = d_p_res.reshape(B, n * n)

            d_phi_pre = d_p_pre.t() @ x_flat
            d_phi_post = d_p_post.t() @ x_flat
            d_phi_res = d_p_res_flat.t() @ x_flat

            d_x_flat = d_p_pre @ phi_pre_f32
            d_x_flat += d_p_post @ phi_post_f32
            d_x_flat += d_p_res_flat @ phi_res_f32

            d_r = -(d_tilde_pre * (alpha_pre_f32 * p_pre) * rms_inv2[:, None]).sum(
                dim=1
            )
            d_r -= (d_tilde_post * (alpha_post_f32 * p_post) * rms_inv2[:, None]).sum(
                dim=1
            )
            d_r -= (d_tilde_res * (alpha_res_f32 * p_res) * rms_inv2.view(B, 1, 1)).sum(
                dim=(1, 2)
            )
            d_x_flat += d_r[:, None] * x_flat * (rms_inv[:, None] / float(nC))

            d_x = d_x + d_x_flat.view(B, n, C)

            d_b_pre = d_b_pre.to(b_pre.dtype)
            d_b_post = d_b_post.to(b_post.dtype)
            d_b_res = d_b_res.to(b_res.dtype)
            d_alpha_pre = d_alpha_pre.to(alpha_pre.dtype)
            d_alpha_post = d_alpha_post.to(alpha_post.dtype)
            d_alpha_res = d_alpha_res.to(alpha_res.dtype)
            d_phi_pre = d_phi_pre.to(phi_pre.dtype)
            d_phi_post = d_phi_post.to(phi_post.dtype)
            d_phi_res = d_phi_res.to(phi_res.dtype)

        return (
            d_x,
            d_rmsnorm_weight,
            d_phi_pre,
            d_phi_post,
            d_phi_res,
            d_alpha_pre,
            d_alpha_post,
            d_alpha_res,
            d_b_pre,
            d_b_post,
            d_b_res,
            None,
            None,
        )


def mhc_layer_fused_dynamic(
    x_expanded,
    rmsnorm_weight,
    phi_pre,
    phi_post,
    phi_res,
    alpha_pre,
    alpha_post,
    alpha_res,
    b_pre,
    b_post,
    b_res,
    sinkhorn_iters=20,
    eps=1e-5,
):
    if not torch.is_tensor(alpha_pre):
        alpha_pre = torch.tensor(alpha_pre, device=phi_pre.device, dtype=phi_pre.dtype)
    if not torch.is_tensor(alpha_post):
        alpha_post = torch.tensor(
            alpha_post, device=phi_pre.device, dtype=phi_pre.dtype
        )
    if not torch.is_tensor(alpha_res):
        alpha_res = torch.tensor(alpha_res, device=phi_pre.device, dtype=phi_pre.dtype)
    return MHCLayerDynamicFunction.apply(
        x_expanded.float(),
        rmsnorm_weight.bfloat16(),
        phi_pre.float(),
        phi_post.float(),
        phi_res.float(),
        alpha_pre,
        alpha_post,
        alpha_res,
        b_pre.float(),
        b_post.float(),
        b_res.float(),
        sinkhorn_iters,
        eps,
    )


def mhc_layer_fused_dynamic_inference(
    x_expanded,
    rmsnorm_weight,
    phi_pre,
    phi_post,
    phi_res,
    alpha_pre,
    alpha_post,
    alpha_res,
    b_pre,
    b_post,
    b_res,
    sinkhorn_iters=20,
    eps=1e-5,
):
    n = phi_pre.size(0)
    phi_res_view = phi_res.view(n * n, -1)
    nC = phi_pre.size(1)
    out_dim = n + n + n * n
    phi_concat = torch.empty(
        (out_dim, nC),
        device=phi_pre.device,
        dtype=torch.bfloat16,
    )
    phi_concat[:n].copy_(phi_pre)
    phi_concat[n : 2 * n].copy_(phi_post)
    phi_concat[2 * n :].copy_(phi_res_view)
    alpha_pre_val = (
        float(alpha_pre.item()) if torch.is_tensor(alpha_pre) else float(alpha_pre)
    )
    alpha_post_val = (
        float(alpha_post.item()) if torch.is_tensor(alpha_post) else float(alpha_post)
    )
    alpha_res_val = (
        float(alpha_res.item()) if torch.is_tensor(alpha_res) else float(alpha_res)
    )
    x_f32 = x_expanded.float().contiguous()
    H_pre_activated, H_post_activated, M, _rms_h = mhc_cuda.mhc_dynamic_h_fwd(
        x_f32,
        phi_concat,
        alpha_pre_val,
        alpha_post_val,
        alpha_res_val,
        b_pre.float().contiguous(),
        b_post.float().contiguous(),
        b_res.float().contiguous(),
        sinkhorn_iters,
        eps,
    )
    x_agg_bf16 = mhc_cuda.mhc_dynamic_aggregate_fwd(x_f32, H_pre_activated)
    y_norm_bf16, _ = mhc_cuda.rmsnorm_fwd(
        x_agg_bf16.contiguous(),
        rmsnorm_weight.bfloat16().contiguous(),
        eps,
    )
    output = mhc_cuda.mhc_dynamic_mix_fwd(
        x_f32,
        y_norm_bf16,
        H_post_activated,
        M,
    )
    return output



from __future__ import annotations

import argparse
import json
import math
import os
import time
from dataclasses import asdict
from itertools import chain
from pathlib import Path
from typing import Optional

import torch
import torch.nn.functional as F
from datasets import load_dataset
from transformers import (
    AutoTokenizer,
    Trainer,
    TrainingArguments,
    default_data_collator,
    set_seed,
)
from transformers.trainer_callback import TrainerCallback
from transformers.utils import SAFE_WEIGHTS_NAME

try:
    from transformers.trainer_utils import TRAINING_ARGS_NAME
except ImportError:  # pragma: no cover - compatibility with older transformers
    try:
        from transformers.utils import TRAINING_ARGS_NAME
    except ImportError:
        TRAINING_ARGS_NAME = "training_args.bin"

from .model import ModelConfig, MHCTransformer, mhc_cuda

MODEL_PRESETS = {
    "3b": {
        "n_layers": 12,
        "hidden_dim": 1280,
        "ffn_dim": 896,
        "n_heads": 16,
        "rope_dim": 64,
        "vocab_size": 129280,
        "max_seq_len": 4096,
    },
    "9b": {
        "n_layers": 18,
        "hidden_dim": 1920,
        "ffn_dim": 1280,
        "n_heads": 24,
        "rope_dim": 64,
        "vocab_size": 129280,
        "max_seq_len": 4096,
    },
    "27b": {
        "n_layers": 30,
        "hidden_dim": 2560,
        "ffn_dim": 1536,
        "n_heads": 32,
        "rope_dim": 64,
        "vocab_size": 129280,
        "max_seq_len": 4096,
    },
    "3b-1t": {
        "n_layers": 12,
        "hidden_dim": 1280,
        "ffn_dim": 896,
        "n_heads": 16,
        "rope_dim": 64,
        "vocab_size": 129280,
        "max_seq_len": 4096,
    },
}

TRAIN_PRESETS = {
    "3b": {
        "batch_size": 320,
        "training_steps": 30000,
        "base_lr": 8.6e-4,
    },
    "9b": {
        "batch_size": 512,
        "training_steps": 50000,
        "base_lr": 5.9e-4,
    },
    "27b": {
        "batch_size": 1280,
        "training_steps": 50000,
        "base_lr": 4.0e-4,
    },
    "3b-1t": {
        "batch_size": 2560,
        "training_steps": 100000,
        "base_lr": 9.0e-4,
    },
}


def _bytes_to_gib(value: int) -> float:
    return value / (1024**3)


def _get_cuda_memory() -> dict[str, float]:
    if not torch.cuda.is_available():
        return {}
    free, total = torch.cuda.mem_get_info()
    return {
        "cuda_mem_allocated_gb": _bytes_to_gib(torch.cuda.memory_allocated()),
        "cuda_mem_reserved_gb": _bytes_to_gib(torch.cuda.memory_reserved()),
        "cuda_mem_peak_gb": _bytes_to_gib(torch.cuda.max_memory_allocated()),
        "cuda_mem_peak_reserved_gb": _bytes_to_gib(torch.cuda.max_memory_reserved()),
        "cuda_mem_free_gb": _bytes_to_gib(free),
        "cuda_mem_total_gb": _bytes_to_gib(total),
    }


def _optimizer_state_gb(optimizer: torch.optim.Optimizer) -> float:
    total_bytes = 0
    for state in optimizer.state.values():
        for value in state.values():
            if torch.is_tensor(value):
                total_bytes += value.numel() * value.element_size()
    return _bytes_to_gib(total_bytes)


def _param_grad_gb(model: torch.nn.Module) -> tuple[float, float]:
    param_bytes = 0
    grad_bytes = 0
    for param in model.parameters():
        param_bytes += param.numel() * param.element_size()
        if param.grad is not None:
            grad_bytes += param.grad.numel() * param.grad.element_size()
    return _bytes_to_gib(param_bytes), _bytes_to_gib(grad_bytes)


def _count_params(model: torch.nn.Module) -> int:
    return sum(p.numel() for p in model.parameters())


def _resolve_text_field(example: dict) -> str:
    for key in ("text", "content", "document", "doc", "data"):
        if key in example:
            return key
    for key, value in example.items():
        if isinstance(value, str):
            return key
    raise ValueError("Unable to infer text field from dataset samples.")


def _prepare_hf_dataset(dataset, tokenizer, text_field: str, seq_len: int, stride: int):
    eos_token_id = tokenizer.eos_token_id

    def tokenize(batch: dict) -> dict:
        encoded = tokenizer(
            batch[text_field],
            add_special_tokens=False,
            return_attention_mask=False,
        )
        if eos_token_id is not None:
            encoded["input_ids"] = [
                ids + [eos_token_id] for ids in encoded["input_ids"]
            ]
        return encoded

    dataset = dataset.map(
        tokenize,
        batched=True,
        remove_columns=dataset.column_names,
    )

    block_size = seq_len + 1
    step = stride or seq_len
    if step <= 0:
        step = seq_len

    def group_texts(batch: dict) -> dict:
        concatenated = list(chain.from_iterable(batch["input_ids"]))
        total_length = len(concatenated)
        if total_length < block_size:
            return {"input_ids": [], "labels": []}
        input_ids = []
        labels = []
        for i in range(0, total_length - block_size + 1, step):
            window = concatenated[i : i + block_size]
            input_ids.append(window[:-1])
            labels.append(window[1:])
        return {"input_ids": input_ids, "labels": labels}

    dataset = dataset.map(group_texts, batched=True)
    return dataset


def _nonempty_example(example: dict) -> bool:
    return len(example["input_ids"]) > 0


def _apply_scale(config: dict, scale: float) -> dict:
    if scale == 1.0:
        return config
    hidden_dim = max(16, int(config["hidden_dim"] * scale))
    n_heads = max(1, int(config["n_heads"] * scale))
    # Ensure head_dim is even (required by RoPE rotate_half)
    head_dim = hidden_dim // n_heads
    if head_dim % 2 != 0:
        head_dim += 1
    hidden_dim = n_heads * head_dim
    ffn_dim = max(16, int(config["ffn_dim"] * scale))
    config = dict(config)
    config.update(hidden_dim=hidden_dim, n_heads=n_heads, ffn_dim=ffn_dim)
    return config


def _build_model_config(args, vocab_size: int) -> ModelConfig:
    preset = _apply_scale(MODEL_PRESETS[args.preset], args.scale)
    hidden_dim = args.hidden_dim or preset["hidden_dim"]
    n_heads = args.n_heads or preset["n_heads"]
    if hidden_dim % n_heads != 0:
        hidden_dim = (hidden_dim // n_heads) * n_heads
    return ModelConfig(
        vocab_size=args.vocab_size or vocab_size or preset["vocab_size"],
        n_layers=args.n_layers or preset["n_layers"],
        hidden_dim=hidden_dim,
        n_heads=n_heads,
        ffn_dim=args.ffn_dim or preset["ffn_dim"],
        max_seq_len=args.seq_len or preset["max_seq_len"],
        expansion_rate=args.expansion_rate,
        sinkhorn_iters=args.sinkhorn_iters,
        alpha_init=args.alpha_init,
        rmsnorm_eps=args.rmsnorm_eps,
        sinkhorn_eps=args.sinkhorn_eps,
        rope_dim=args.rope_dim or preset["rope_dim"],
        rope_theta=args.rope_theta,
        dropout=args.dropout,
        sdp_kernel=args.sdp_kernel,
        recompute_ratio=args.recompute_ratio,
        recompute_mhc=args.recompute_mhc,
        use_dynamic_h=args.dynamic_h,
        mlp_type=args.mlp_type,
    )


class _MetricsCallback(TrainerCallback):
    def __init__(
        self,
        metrics_path: Path,
        log_memory: bool,
        tokens_per_step: int,
    ) -> None:
        self.metrics_path = metrics_path
        self.log_memory = log_memory
        self.tokens_per_step = tokens_per_step
        self.start_time: Optional[float] = None

    def on_train_begin(self, args, state, control, **kwargs):
        self.start_time = time.time()

    def on_log(self, args, state, control, logs=None, **kwargs):
        if logs is None:
            return
        log = dict(logs)
        if self.start_time is not None:
            elapsed = time.time() - self.start_time
            tokens_seen = state.global_step * self.tokens_per_step
            log["tokens_seen"] = tokens_seen
            log["tokens_per_s"] = tokens_seen / max(elapsed, 1e-6)
        log_cuda = self.log_memory and torch.cuda.is_available()
        if log_cuda:
            log.update(_get_cuda_memory())
            model = kwargs.get("model")
            optimizer = kwargs.get("optimizer")
            if model is not None:
                param_gb, grad_gb = _param_grad_gb(model)
                log["param_gb"] = param_gb
                log["grad_gb"] = grad_gb
            if optimizer is not None:
                log["optim_state_gb"] = _optimizer_state_gb(optimizer)
        lr = log.get("learning_rate")
        loss = log.get("loss")
        tokens_per_s = log.get("tokens_per_s")
        if loss is not None and lr is not None and tokens_per_s is not None:
            print(
                f"step={state.global_step} loss={loss:.4f} lr={lr:.3e} "
                f"tokens/s={tokens_per_s:.1f}"
            )
        if log_cuda and "cuda_mem_allocated_gb" in log:
            print(
                "cuda_mem="
                f"{log['cuda_mem_allocated_gb']:.1f}G/"
                f"{log['cuda_mem_reserved_gb']:.1f}G "
                f"peak={log['cuda_mem_peak_gb']:.1f}G"
            )
            if "param_gb" in log and "optim_state_gb" in log:
                print(
                    "mem_breakdown="
                    f"params={log.get('param_gb', 0.0):.1f}G "
                    f"grads={log.get('grad_gb', 0.0):.1f}G "
                    f"optim={log.get('optim_state_gb', 0.0):.1f}G"
                )
        with self.metrics_path.open("a") as f:
            f.write(json.dumps(log) + "\n")


class MHCTrainer(Trainer):
    def __init__(
        self,
        *args,
        logits_chunk_size: int,
        loss_dtype: str,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.logits_chunk_size = logits_chunk_size
        self.loss_dtype = loss_dtype

    def _safe_state_dict(self, model: torch.nn.Module) -> dict[str, torch.Tensor]:
        state = model.state_dict()
        token_key = "token_emb.weight"
        head_key = "lm_head.weight"
        if token_key in state and head_key in state:
            if state[token_key].data_ptr() == state[head_key].data_ptr():
                state[head_key] = state[head_key].clone()
        safe_state: dict[str, torch.Tensor] = {}
        for key, value in state.items():
            if torch.is_tensor(value):
                tensor = value.detach().cpu()
                if not tensor.is_contiguous():
                    tensor = tensor.contiguous()
                safe_state[key] = tensor
        return safe_state

    def save_model(
        self, output_dir: Optional[str] = None, _internal_call: bool = False
    ) -> None:
        if not self.is_world_process_zero():
            return
        if output_dir is None:
            output_dir = self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        state_dict = self._safe_state_dict(self.model)
        from safetensors.torch import save_file

        save_file(
            state_dict,
            os.path.join(output_dir, SAFE_WEIGHTS_NAME),
            metadata={"format": "pt"},
        )
        torch.save(self.args, os.path.join(output_dir, TRAINING_ARGS_NAME))
        if hasattr(self.model, "config"):
            config_path = os.path.join(output_dir, "model_config.json")
            with open(config_path, "w") as f:
                json.dump(asdict(self.model.config), f, indent=2)

    def _resolve_inputs(
        self, inputs: dict[str, torch.Tensor]
    ) -> tuple[torch.Tensor, torch.Tensor]:
        input_ids = inputs.get("input_ids")
        labels = inputs.get("labels")
        if labels is None:
            labels = inputs.get("label")
        if labels is None and input_ids is not None:
            labels = input_ids.clone()
            if labels.dim() == 2 and labels.size(1) > 1:
                labels[:, :-1] = input_ids[:, 1:]
                labels[:, -1] = -100
        if input_ids is None or labels is None:
            raise ValueError("inputs must contain input_ids and labels")
        return input_ids, labels

    def _chunked_ce_loss(
        self, hidden: torch.Tensor, labels: torch.Tensor, lm_head: torch.nn.Module
    ) -> float:
        hidden_flat = hidden.reshape(-1, hidden.size(-1))
        labels_flat = labels.reshape(-1)
        num_positions = labels_flat.numel()
        valid_tokens = (labels_flat != -100).sum().item()
        if valid_tokens == 0:
            return 0.0
        total_loss_val = 0.0
        head_dtype = lm_head.weight.dtype
        grad_accum = self.args.gradient_accumulation_steps
        for start in range(0, num_positions, self.logits_chunk_size):
            end = min(start + self.logits_chunk_size, num_positions)
            chunk = hidden_flat[start:end]
            if chunk.dtype != head_dtype:
                chunk = chunk.to(head_dtype)
            logits = lm_head(chunk)
            logits_for_loss = logits.float() if self.loss_dtype == "fp32" else logits
            chunk_loss = F.cross_entropy(
                logits_for_loss,
                labels_flat[start:end],
                reduction="sum",
                ignore_index=-100,
            )
            total_loss_val += chunk_loss.item()
            chunk_loss = chunk_loss / float(valid_tokens * grad_accum)
            self.accelerator.backward(chunk_loss, retain_graph=end < num_positions)
            del logits, chunk_loss
        return total_loss_val / max(valid_tokens, 1)

    def _training_step_chunked(
        self, model: torch.nn.Module, input_ids: torch.Tensor, labels: torch.Tensor
    ) -> torch.Tensor:
        hidden = model(input_ids, return_hidden=True)
        loss_val = self._chunked_ce_loss(hidden, labels, model.lm_head)
        loss_val = loss_val / float(self.args.gradient_accumulation_steps)
        return hidden.new_tensor(loss_val, dtype=torch.float32)

    def _training_step_full(
        self, model: torch.nn.Module, input_ids: torch.Tensor, labels: torch.Tensor
    ) -> torch.Tensor:
        logits = model(input_ids)
        logits_for_loss = logits.float() if self.loss_dtype == "fp32" else logits
        loss = F.cross_entropy(
            logits_for_loss.view(-1, logits_for_loss.size(-1)),
            labels.view(-1),
            ignore_index=-100,
        )
        loss = loss / self.args.gradient_accumulation_steps
        self.accelerator.backward(loss)
        return loss

    def training_step(self, model, inputs, _num_items_in_batch=None):
        model.train()
        inputs = self._prepare_inputs(inputs)
        input_ids, labels = self._resolve_inputs(inputs)

        with self.compute_loss_context_manager():
            if self.logits_chunk_size > 0:
                loss = self._training_step_chunked(model, input_ids, labels)
            else:
                loss = self._training_step_full(model, input_ids, labels)

        return loss.detach()


def run_training(argv: Optional[list[str]] = None):
    parser = argparse.ArgumentParser(description="mHC trainer")
    parser.add_argument("--preset", default="3b", choices=sorted(MODEL_PRESETS))
    parser.add_argument("--scale", type=float, default=1.0)
    parser.add_argument("--dataset", default="Trelis/tiny-shakespeare")
    parser.add_argument("--dataset-split", default="train")
    parser.add_argument("--eval-dataset", default=None)
    parser.add_argument("--eval-split", default="validation")
    parser.add_argument("--eval-steps", type=int, default=0)
    parser.add_argument("--text-field", default=None)
    parser.add_argument("--tokenizer", default="deepseek-ai/DeepSeek-V3")
    parser.add_argument("--seq-len", type=int, default=None)
    parser.add_argument("--stride", type=int, default=None)
    parser.add_argument("--batch-size", type=int, default=None)
    parser.add_argument("--grad-accum", type=int, default=1)
    parser.add_argument("--max-steps", type=int, default=None)
    parser.add_argument("--warmup-steps", type=int, default=None)
    parser.add_argument("--lr", type=float, default=None)
    parser.add_argument(
        "--lr-schedule",
        choices=("cosine", "linear"),
        default="cosine",
        help="LR schedule after warmup. Default: cosine (matches paper).",
    )
    parser.add_argument(
        "--lr-scale",
        choices=("none", "linear", "sqrt"),
        default="linear",
        help="Scale preset LR by effective_batch / preset_batch when --lr is unset.",
    )
    parser.add_argument("--weight-decay", type=float, default=0.1)
    parser.add_argument("--beta1", type=float, default=0.9)
    parser.add_argument("--beta2", type=float, default=0.95)
    parser.add_argument("--adam-eps", type=float, default=1e-8)
    parser.add_argument("--grad-clip", type=float, default=1.0)
    parser.add_argument("--expansion-rate", type=int, default=4)
    parser.add_argument("--sinkhorn-iters", type=int, default=20)
    parser.add_argument("--alpha-init", type=float, default=0.01)
    parser.add_argument("--rmsnorm-eps", type=float, default=1e-6)
    parser.add_argument("--sinkhorn-eps", type=float, default=1e-6)
    parser.add_argument("--rope-dim", type=int, default=None)
    parser.add_argument("--rope-theta", type=float, default=10000.0)
    parser.add_argument("--dropout", type=float, default=0.0)
    parser.add_argument(
        "--sdp-kernel",
        choices=("auto", "flash", "math"),
        default="flash",
        help="Scaled dot-product attention kernel selection.",
    )
    parser.add_argument(
        "--recompute-ratio",
        type=float,
        default=0.0,
        help="Fraction of transformer blocks to checkpoint (0 disables, 1 checkpoints all).",
    )
    parser.add_argument(
        "--recompute-mhc",
        action="store_true",
        help="Recompute residual-stream kernels in backward without rerunning attention/MLP.",
    )
    parser.add_argument(
        "--dynamic-h",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Use dynamic (input-dependent) H matrices. Default: True (matches paper).",
    )
    parser.add_argument(
        "--mlp-type",
        choices=("swiglu", "gelu"),
        default="swiglu",
        help="MLP type. Default: swiglu (matches DeepSeek architecture).",
    )
    parser.add_argument("--n-layers", type=int, default=None)
    parser.add_argument("--hidden-dim", type=int, default=None)
    parser.add_argument("--n-heads", type=int, default=None)
    parser.add_argument("--ffn-dim", type=int, default=None)
    parser.add_argument("--vocab-size", type=int, default=None)
    parser.add_argument("--dtype", choices=("fp32", "bf16"), default="bf16")
    parser.add_argument(
        "--loss-dtype",
        choices=("bf16", "fp32"),
        default="fp32",
        help="Compute LM head + loss in this dtype for stability.",
    )
    parser.add_argument("--seed", type=int, default=1234)
    parser.add_argument("--log-every", type=int, default=10)
    parser.add_argument("--save-every", type=int, default=200)
    parser.add_argument(
        "--log-memory",
        action="store_true",
        help="Log CUDA memory stats alongside loss.",
    )
    parser.add_argument(
        "--logits-chunk-size",
        type=int,
        default=0,
        help="Compute LM head + loss in token chunks to reduce peak memory.",
    )
    parser.add_argument(
        "--no-fused",
        action="store_true",
        help="Disable fused mHC CUDA kernels (uses pure PyTorch path).",
    )
    parser.add_argument(
        "--torch-compile",
        action="store_true",
        help="Apply torch.compile(mode='reduce-overhead') for CUDA graph optimization.",
    )
    parser.add_argument("--runs-root", default=os.getenv("MHC_RUNS_ROOT", "runs"))
    parser.add_argument("--run-name", default=None)
    parser.add_argument("--num-workers", type=int, default=4)
    parser.add_argument("--shuffle-buffer", type=int, default=1000)
    parser.add_argument(
        "--streaming",
        action="store_true",
        help="Stream dataset samples instead of fully caching to disk.",
    )
    parser.add_argument("--hf-token", default=os.getenv("HF_TOKEN"))
    args = parser.parse_args(argv)
    if args.logits_chunk_size < 0:
        raise ValueError("logits_chunk_size must be >= 0")

    preset_train = TRAIN_PRESETS[args.preset]
    if args.batch_size is None:
        args.batch_size = preset_train["batch_size"]
    if args.max_steps is None:
        args.max_steps = preset_train["training_steps"]
    effective_batch = args.batch_size * args.grad_accum

    base_lr = args.lr or preset_train["base_lr"]
    lr_scale_factor = 1.0
    if args.lr is None and args.lr_scale != "none":
        lr_scale_factor = effective_batch / preset_train["batch_size"]
        if args.lr_scale == "sqrt":
            lr_scale_factor = math.sqrt(lr_scale_factor)
        base_lr *= lr_scale_factor

    warmup_steps = args.warmup_steps or min(2000, args.max_steps)
    use_fused_mhc = not args.no_fused
    if use_fused_mhc and mhc_cuda is None:
        print("mhc_cuda not found; falling back to python mHC path.")
        use_fused_mhc = False
    print(
        f"mhc_cuda: {'available' if mhc_cuda is not None else 'NOT available'}, use_fused_mhc={use_fused_mhc}"
    )
    if use_fused_mhc and args.sinkhorn_eps != args.rmsnorm_eps:
        fused_eps = max(args.sinkhorn_eps, args.rmsnorm_eps)
        print(f"Fused mHC uses a single eps; using eps={fused_eps:g}.")
        args.sinkhorn_eps = fused_eps
        args.rmsnorm_eps = fused_eps

    set_seed(args.seed)
    torch.backends.cuda.matmul.allow_tf32 = True

    tokenizer = AutoTokenizer.from_pretrained(
        args.tokenizer, use_fast=True, token=args.hf_token
    )
    if tokenizer.pad_token is None:
        tokenizer.pad_token = tokenizer.eos_token

    dataset = load_dataset(
        args.dataset,
        split=args.dataset_split,
        streaming=args.streaming,
        token=args.hf_token,
    )
    sample = next(iter(dataset)) if args.streaming else dataset[0]
    text_field = args.text_field or _resolve_text_field(sample)

    model_config = _build_model_config(args, tokenizer.vocab_size)
    model_config.use_fused_mhc = use_fused_mhc
    model = MHCTransformer(model_config)
    if args.dtype == "bf16":
        model = model.to(dtype=torch.bfloat16)

    if args.torch_compile:
        model = torch.compile(model, mode="max-autotune", dynamic=False)

    n_params = _count_params(model)
    print(f"model params: {n_params:,} ({n_params / 1e6:.1f}M)")

    seq_len = model_config.max_seq_len
    stride = args.stride or seq_len
    dataset = _prepare_hf_dataset(dataset, tokenizer, text_field, seq_len, stride)
    dataset = dataset.filter(_nonempty_example)
    if args.shuffle_buffer:
        if args.streaming:
            dataset = dataset.shuffle(buffer_size=args.shuffle_buffer, seed=args.seed)
        else:
            dataset = dataset.shuffle(seed=args.seed)
    dataset = dataset.with_format("torch")

    eval_dataset = None
    if args.eval_dataset or args.eval_steps > 0:
        eval_source = args.eval_dataset or args.dataset
        eval_ds = load_dataset(
            eval_source,
            split=args.eval_split,
            streaming=args.streaming,
            token=args.hf_token,
        )
        eval_ds = _prepare_hf_dataset(eval_ds, tokenizer, text_field, seq_len, stride)
        eval_ds = eval_ds.filter(_nonempty_example)
        eval_dataset = eval_ds.with_format("torch")

    run_name = args.run_name or f"mhc-{args.preset}-{int(time.time())}"
    out_dir = Path(args.runs_root) / run_name
    out_dir.mkdir(parents=True, exist_ok=True)

    with (out_dir / "config.json").open("w") as f:
        json.dump(
            {
                "preset": args.preset,
                "preset_batch_size": preset_train["batch_size"],
                "preset_training_steps": preset_train["training_steps"],
                "base_lr": base_lr,
                "lr_schedule": args.lr_schedule,
                "lr_scale": args.lr_scale,
                "lr_scale_factor": lr_scale_factor,
                "warmup_steps": warmup_steps,
                "dataset": args.dataset,
                "tokenizer": args.tokenizer,
                "text_field": text_field,
                "model_config": asdict(model_config),
                "n_params": n_params,
                "batch_size": args.batch_size,
                "grad_accum": args.grad_accum,
                "effective_batch_size": effective_batch,
                "max_steps": args.max_steps,
                "grad_clip": args.grad_clip,
                "sdp_kernel": args.sdp_kernel,
                "recompute_ratio": args.recompute_ratio,
                "recompute_mhc": args.recompute_mhc,
                "logits_chunk_size": args.logits_chunk_size,
                "use_fused_mhc": use_fused_mhc,
            },
            f,
            indent=2,
        )

    with (out_dir / "model_config.json").open("w") as f:
        json.dump(asdict(model_config), f, indent=2)

    if args.log_memory and torch.cuda.is_available():
        param_gb, _ = _param_grad_gb(model)
        print(f"model_params={param_gb:.2f}G (bf16 storage)")

    eval_strategy = "no"
    eval_steps_val = None
    if eval_dataset is not None and args.eval_steps > 0:
        eval_strategy = "steps"
        eval_steps_val = args.eval_steps

    ta_kwargs = dict(
        output_dir=str(out_dir),
        per_device_train_batch_size=args.batch_size,
        gradient_accumulation_steps=args.grad_accum,
        max_steps=args.max_steps,
        warmup_steps=warmup_steps,
        learning_rate=base_lr,
        lr_scheduler_type=args.lr_schedule,
        weight_decay=args.weight_decay,
        adam_beta1=args.beta1,
        adam_beta2=args.beta2,
        adam_epsilon=args.adam_eps,
        max_grad_norm=args.grad_clip,
        bf16=args.dtype == "bf16",
        logging_steps=args.log_every,
        logging_first_step=True,
        save_steps=args.save_every,
        save_strategy="steps" if args.save_every > 0 else "no",
        logging_strategy="steps",
        report_to=[],
        dataloader_num_workers=args.num_workers,
        dataloader_pin_memory=torch.cuda.is_available(),
        run_name=run_name,
        remove_unused_columns=False,
    )
    import inspect

    _ta_params = inspect.signature(TrainingArguments).parameters
    # eval_strategy was renamed from evaluation_strategy in newer transformers
    if eval_strategy != "no":
        if "eval_strategy" in _ta_params:
            ta_kwargs["eval_strategy"] = eval_strategy
        else:
            ta_kwargs["evaluation_strategy"] = eval_strategy
        ta_kwargs["eval_steps"] = eval_steps_val
    # save_safetensors added in transformers >= 4.36
    if "save_safetensors" in _ta_params:
        ta_kwargs["save_safetensors"] = True
    training_args = TrainingArguments(**ta_kwargs)

    tokens_per_step = args.batch_size * args.grad_accum * seq_len
    metrics_path = out_dir / "metrics.jsonl"

    trainer = MHCTrainer(
        model=model,
        args=training_args,
        train_dataset=dataset,
        eval_dataset=eval_dataset,
        data_collator=default_data_collator,
        logits_chunk_size=args.logits_chunk_size,
        loss_dtype=args.loss_dtype,
    )
    trainer.add_callback(
        _MetricsCallback(
            metrics_path=metrics_path,
            log_memory=args.log_memory,
            tokens_per_step=tokens_per_step,
        )
    )

    trainer.train()
    trainer.save_model(str(out_dir))
    return run_name, str(out_dir)


if __name__ == "__main__":
    run_training()



import torch
import torch.nn as nn

from .ops import (
    mhc_layer_fused,
    mhc_layer_fused_dynamic,
    mhc_layer_fused_dynamic_inference,
    mhc_layer_fused_inference,
)


class MHCLayer(nn.Module):
    """
    Args:
        hidden_dim: The hidden dimension (C).
        expansion_rate: The expansion rate (n).
        sinkhorn_iters: Number of Sinkhorn-Knopp iterations.
        eps: Epsilon for numerical stability.
        alpha_init: Initialization scale for alpha parameters.
        use_dynamic_h: If True, uses per-batch H values computed from x.
                       If False, uses static shared H values.
    """

    def __init__(
        self,
        hidden_dim: int,
        expansion_rate: int = 4,
        sinkhorn_iters: int = 20,
        eps: float = 1e-5,
        alpha_init: float = 0.01,
        use_dynamic_h: bool = True,
    ):
        super().__init__()
        self.hidden_dim = hidden_dim
        self.expansion_rate = expansion_rate
        self.sinkhorn_iters = sinkhorn_iters
        self.eps = eps
        self.use_dynamic_h = use_dynamic_h

        n = expansion_rate
        C = hidden_dim
        nC = n * C

        self.rmsnorm_weight = nn.Parameter(torch.ones(hidden_dim, dtype=torch.bfloat16))

        if use_dynamic_h:
            self.phi_pre = nn.Parameter(torch.randn(n, nC) * 0.02)
            self.phi_post = nn.Parameter(torch.randn(n, nC) * 0.02)
            self.phi_res = nn.Parameter(torch.randn(n * n, nC) * 0.02)

            self.b_pre = nn.Parameter(torch.zeros(n))
            self.b_post = nn.Parameter(torch.zeros(n))
            self.b_res = nn.Parameter(torch.zeros(n, n))

            self.alpha_pre = nn.Parameter(torch.tensor(alpha_init))
            self.alpha_post = nn.Parameter(torch.tensor(alpha_init))
            self.alpha_res = nn.Parameter(torch.tensor(alpha_init))
        else:
            self.H_pre = nn.Parameter(torch.zeros(expansion_rate, dtype=torch.float32))
            self.H_post = nn.Parameter(torch.zeros(expansion_rate, dtype=torch.float32))
            H_res_init = alpha_init * torch.randn(expansion_rate, expansion_rate)
            self.H_res = nn.Parameter(H_res_init.float())

    def forward(self, x_expanded: torch.Tensor) -> torch.Tensor:
        B, n, C = x_expanded.shape
        assert n == self.expansion_rate
        assert C == self.hidden_dim

        if self.use_dynamic_h:
            if not self.training and not torch.is_grad_enabled():
                return mhc_layer_fused_dynamic_inference(
                    x_expanded,
                    self.rmsnorm_weight,
                    self.phi_pre,
                    self.phi_post,
                    self.phi_res,
                    self.alpha_pre,
                    self.alpha_post,
                    self.alpha_res,
                    self.b_pre,
                    self.b_post,
                    self.b_res,
                    self.sinkhorn_iters,
                    self.eps,
                )
            return mhc_layer_fused_dynamic(
                x_expanded,
                self.rmsnorm_weight,
                self.phi_pre,
                self.phi_post,
                self.phi_res,
                self.alpha_pre,
                self.alpha_post,
                self.alpha_res,
                self.b_pre,
                self.b_post,
                self.b_res,
                self.sinkhorn_iters,
                self.eps,
            )
        else:
            if not self.training and not torch.is_grad_enabled():
                return mhc_layer_fused_inference(
                    x_expanded,
                    self.rmsnorm_weight,
                    self.H_pre,
                    self.H_post,
                    self.H_res,
                    self.sinkhorn_iters,
                    self.eps,
                )
            return mhc_layer_fused(
                x_expanded,
                self.rmsnorm_weight,
                self.H_pre,
                self.H_post,
                self.H_res,
                self.sinkhorn_iters,
                self.eps,
            )



from __future__ import annotations

import contextlib
from dataclasses import dataclass

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function
from torch.utils.checkpoint import checkpoint

try:
    import mhc_cuda
except ImportError:  # pragma: no cover - optional fused path
    mhc_cuda = None

try:
    from . import ops as mhc_ops
except ImportError:  # pragma: no cover - optional fused path
    mhc_ops = None


@dataclass
class ModelConfig:
    vocab_size: int
    n_layers: int
    hidden_dim: int
    n_heads: int
    ffn_dim: int
    max_seq_len: int = 4096
    expansion_rate: int = 4
    sinkhorn_iters: int = 20
    alpha_init: float = 0.01
    rmsnorm_eps: float = 1e-5
    sinkhorn_eps: float = 1e-6
    rope_dim: int = 64
    rope_theta: float = 10000.0
    dropout: float = 0.0
    sdp_kernel: str = "flash"
    recompute_ratio: float = 0.0
    recompute_mhc: bool = False
    use_dynamic_h: bool = True
    use_fused_mhc: bool = True
    mlp_type: str = "swiglu"


H_RES_EXP_CLAMP = 20.0


_SDP_KERNEL_SETTINGS = {
    "auto": (True, True, True),
    "flash": (True, False, False),
    "mem-efficient": (False, True, False),
    "math": (False, False, True),
}


def _sdp_kernel_context(kernel: str, device: torch.device):
    if kernel == "auto" or device.type != "cuda":
        return contextlib.nullcontext()
    if kernel not in _SDP_KERNEL_SETTINGS:
        raise ValueError(f"Unsupported sdp kernel: {kernel}")
    enable_flash, enable_mem_efficient, enable_math = _SDP_KERNEL_SETTINGS[kernel]
    attention_mod = getattr(torch.nn, "attention", None)
    sdp_kernel = getattr(attention_mod, "sdpa_kernel", None) if attention_mod else None
    if sdp_kernel is not None:
        sdp_backend = getattr(attention_mod, "SDPBackend", None)
        if sdp_backend is not None:
            backend_map = {
                "flash": getattr(sdp_backend, "FLASH_ATTENTION", None),
                "mem-efficient": getattr(sdp_backend, "EFFICIENT_ATTENTION", None)
                or getattr(sdp_backend, "MEM_EFFICIENT", None),
                "math": getattr(sdp_backend, "MATH", None),
            }
            backend = backend_map[kernel]
            if backend is not None:
                for candidate in (backend, [backend], {backend}):
                    try:
                        return sdp_kernel(candidate)
                    except TypeError:
                        continue
        try:
            return sdp_kernel(
                enable_flash=enable_flash,
                enable_mem_efficient=enable_mem_efficient,
                enable_math=enable_math,
            )
        except TypeError:
            return contextlib.nullcontext()
    if hasattr(torch.backends.cuda, "sdp_kernel"):
        return torch.backends.cuda.sdp_kernel(
            enable_flash=enable_flash,
            enable_mem_efficient=enable_mem_efficient,
            enable_math=enable_math,
        )
    return contextlib.nullcontext()


def _checkpoint_block(block: nn.Module, x: torch.Tensor) -> torch.Tensor:
    try:
        return checkpoint(block, x, use_reentrant=False, preserve_rng_state=False)
    except TypeError:
        try:
            return checkpoint(block, x, use_reentrant=False)
        except TypeError:
            return checkpoint(block, x)


def _checkpoint_fn(fn, *args):
    try:
        return checkpoint(fn, *args, use_reentrant=False, preserve_rng_state=False)
    except TypeError:
        try:
            return checkpoint(fn, *args, use_reentrant=False)
        except TypeError:
            return checkpoint(fn, *args)


class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float) -> None:
        super().__init__()
        self.weight = nn.Parameter(torch.ones(dim))
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        rms = x.pow(2).mean(dim=-1, keepdim=True).add(self.eps).sqrt()
        return x / rms * self.weight


class DynamicHFusedFunction(Function):
    @staticmethod
    def forward(
        ctx,
        x_expanded: torch.Tensor,
        phi_pre: torch.Tensor,
        phi_post: torch.Tensor,
        phi_res: torch.Tensor,
        alpha_pre: torch.Tensor,
        alpha_post: torch.Tensor,
        alpha_res: torch.Tensor,
        b_pre: torch.Tensor,
        b_post: torch.Tensor,
        b_res: torch.Tensor,
        sinkhorn_iters: int,
        eps: float,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        if mhc_cuda is None:
            raise RuntimeError("mhc_cuda is required for fused mHC")
        n = phi_pre.size(0)
        phi_concat = (
            torch.cat([phi_pre, phi_post, phi_res.view(n * n, -1)], dim=0)
            .bfloat16()
            .contiguous()
        )
        alpha_pre_val = float(alpha_pre.item())
        alpha_post_val = float(alpha_post.item())
        alpha_res_val = float(alpha_res.item())
        b_pre_f32 = b_pre.float().contiguous()
        b_post_f32 = b_post.float().contiguous()
        b_res_f32 = b_res.float().contiguous()
        H_pre_activated, H_post_activated, M, rms_h = mhc_cuda.mhc_dynamic_h_fwd(
            x_expanded.contiguous(),
            phi_concat,
            alpha_pre_val,
            alpha_post_val,
            alpha_res_val,
            b_pre_f32,
            b_post_f32,
            b_res_f32,
            sinkhorn_iters,
            eps,
        )
        ctx.save_for_backward(
            x_expanded,
            phi_pre,
            phi_post,
            phi_res,
            b_pre,
            b_post,
            b_res,
            H_pre_activated,
            H_post_activated,
            M,
            rms_h,
        )
        # Cache scalar float values to avoid .item() sync in backward
        ctx.alpha_pre_val = alpha_pre_val
        ctx.alpha_post_val = alpha_post_val
        ctx.alpha_res_val = alpha_res_val
        ctx.alpha_dtype = alpha_pre.dtype
        ctx.sinkhorn_iters = sinkhorn_iters
        ctx.eps = eps
        return H_pre_activated, H_post_activated, M

    @staticmethod
    def backward(ctx, d_H_pre: torch.Tensor, d_H_post: torch.Tensor, d_M: torch.Tensor):
        (
            x_expanded,
            phi_pre,
            phi_post,
            phi_res,
            b_pre,
            b_post,
            b_res,
            H_pre_activated,
            H_post_activated,
            M,
            rms_h,
        ) = ctx.saved_tensors

        # Use cached float values — no .item() sync needed
        alpha_pre_val = ctx.alpha_pre_val
        alpha_post_val = ctx.alpha_post_val
        alpha_res_val = ctx.alpha_res_val

        B, n, C = x_expanded.shape
        nC = n * C
        x_flat = x_expanded.reshape(B, nC).float()
        rms_inv = (1.0 / rms_h).contiguous()

        phi_pre_f32 = phi_pre.float()
        phi_post_f32 = phi_post.float()
        phi_res_f32 = phi_res.float()

        # Fused projection: single GEMM instead of 3
        phi_concat_f32 = torch.cat(
            [phi_pre_f32, phi_post_f32, phi_res_f32.view(n * n, nC)], dim=0
        )
        p_concat = x_flat @ phi_concat_f32.t()

        # Fused pre-sinkhorn kernel: sigmoid derivatives + H_res_exp recomputation
        d_tilde_pre, d_tilde_post, H_res_exp = mhc_cuda.h_backward_pre(
            d_H_pre.contiguous(),
            d_H_post.contiguous(),
            H_pre_activated.contiguous(),
            H_post_activated.contiguous(),
            p_concat.contiguous(),
            rms_inv,
            alpha_res_val,
            b_res.float().contiguous(),
            n,
        )

        # Sinkhorn backward (unchanged)
        d_H_res_exp = mhc_cuda.sinkhorn_knopp_bwd_batched(
            d_M.contiguous(),
            H_res_exp.contiguous(),
            ctx.sinkhorn_iters,
            ctx.eps,
        )

        # Fused post-sinkhorn kernel: all parameter grads + d_p + d_r
        (
            d_p_pre,
            d_p_post,
            d_p_res_flat,
            d_b_pre,
            d_b_post,
            d_b_res,
            d_alpha_pre,
            d_alpha_post,
            d_alpha_res,
            d_r,
        ) = mhc_cuda.h_backward_post(
            d_H_res_exp.contiguous(),
            H_res_exp.contiguous(),
            d_tilde_pre.contiguous(),
            d_tilde_post.contiguous(),
            p_concat.contiguous(),
            rms_inv,
            alpha_pre_val,
            alpha_post_val,
            alpha_res_val,
            n,
        )

        # Fused weight gradient: single GEMM instead of 3
        d_p_concat = torch.cat([d_p_pre, d_p_post, d_p_res_flat], dim=1)
        d_phi_concat = d_p_concat.t() @ x_flat
        d_phi_pre = d_phi_concat[:n]
        d_phi_post = d_phi_concat[n : 2 * n]
        d_phi_res = d_phi_concat[2 * n :]

        # Fused input gradient: single GEMM instead of 3
        d_x_flat = d_p_concat @ phi_concat_f32

        # Fused RMS correction kernel
        mhc_cuda.rms_correction(d_x_flat.contiguous(), d_r, x_flat, rms_inv, nC)

        d_x = d_x_flat.view(B, n, C).to(x_expanded.dtype)

        # Squeeze scalar outputs from post-sinkhorn kernel
        d_alpha_pre = d_alpha_pre.squeeze()
        d_alpha_post = d_alpha_post.squeeze()
        d_alpha_res = d_alpha_res.squeeze()

        alpha_dtype = ctx.alpha_dtype
        return (
            d_x,
            d_phi_pre.to(phi_pre.dtype),
            d_phi_post.to(phi_post.dtype),
            d_phi_res.view_as(phi_res).to(phi_res.dtype),
            d_alpha_pre.to(alpha_dtype),
            d_alpha_post.to(alpha_dtype),
            d_alpha_res.to(alpha_dtype),
            d_b_pre.to(b_pre.dtype),
            d_b_post.to(b_post.dtype),
            d_b_res.to(b_res.dtype),
            None,
            None,
        )


class AggregateFusedFunction(Function):
    @staticmethod
    def forward(
        ctx, x_expanded: torch.Tensor, H_pre_activated: torch.Tensor
    ) -> torch.Tensor:
        if mhc_cuda is None:
            raise RuntimeError("mhc_cuda is required for fused mHC")
        x_agg = mhc_cuda.mhc_dynamic_aggregate_fwd(
            x_expanded.contiguous(),
            H_pre_activated.contiguous(),
        )
        ctx.save_for_backward(x_expanded, H_pre_activated)
        return x_agg

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        x_expanded, H_pre_activated = ctx.saved_tensors
        d_x, d_H_pre = mhc_cuda.mhc_dynamic_aggregate_bwd(
            grad_output.contiguous(),
            x_expanded.contiguous(),
            H_pre_activated.contiguous(),
        )
        return d_x, d_H_pre


class MixFusedFunction(Function):
    @staticmethod
    def forward(
        ctx,
        x_expanded: torch.Tensor,
        y_bf16: torch.Tensor,
        H_post_activated: torch.Tensor,
        M: torch.Tensor,
    ) -> torch.Tensor:
        if mhc_cuda is None:
            raise RuntimeError("mhc_cuda is required for fused mHC")
        output = mhc_cuda.mhc_dynamic_mix_fwd(
            x_expanded.contiguous(),
            y_bf16.contiguous(),
            H_post_activated.contiguous(),
            M.contiguous(),
        )
        ctx.save_for_backward(x_expanded, y_bf16, H_post_activated, M)
        return output

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        x_expanded, y_bf16, H_post_activated, M = ctx.saved_tensors
        d_x, d_y, d_H_post, d_M = mhc_cuda.mhc_dynamic_mix_bwd(
            grad_output.contiguous(),
            x_expanded.contiguous(),
            y_bf16.contiguous(),
            H_post_activated.contiguous(),
            M.contiguous(),
        )
        return d_x, d_y, d_H_post, d_M


def _rotate_half(x: torch.Tensor) -> torch.Tensor:
    x1 = x[..., ::2]
    x2 = x[..., 1::2]
    return torch.stack((-x2, x1), dim=-1).reshape_as(x)


class RotaryEmbedding(nn.Module):
    def __init__(self, dim: int, base: float) -> None:
        super().__init__()
        inv_freq = 1.0 / (base ** (torch.arange(0, dim, 2, dtype=torch.float32) / dim))
        self.register_buffer("inv_freq", inv_freq, persistent=False)

    def forward(
        self, seq_len: int, device: torch.device, dtype: torch.dtype
    ) -> tuple[torch.Tensor, torch.Tensor]:
        t = torch.arange(seq_len, device=device, dtype=self.inv_freq.dtype)
        freqs = torch.einsum("i,j->ij", t, self.inv_freq)
        emb = torch.cat([freqs, freqs], dim=-1)
        return emb.cos().to(dtype), emb.sin().to(dtype)


class CausalSelfAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        n_heads: int,
        rope_dim: int,
        rope_theta: float,
        dropout: float,
        sdp_kernel: str = "auto",
    ) -> None:
        super().__init__()
        if dim % n_heads != 0:
            raise ValueError("hidden_dim must be divisible by n_heads")
        self.n_heads = n_heads
        self.head_dim = dim // n_heads
        self.rope_dim = min(rope_dim, self.head_dim)
        # RoPE rotate_half requires even dimension
        self.rope_dim = self.rope_dim - (self.rope_dim % 2)
        self.rope = (
            RotaryEmbedding(self.rope_dim, rope_theta) if self.rope_dim > 0 else None
        )
        self.qkv = nn.Linear(dim, dim * 3, bias=False)
        self.proj = nn.Linear(dim, dim, bias=False)
        self.dropout = dropout
        self.sdp_kernel = sdp_kernel

    def _apply_rope(
        self, x: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> torch.Tensor:
        x_rope = x[..., : self.rope_dim]
        x_pass = x[..., self.rope_dim :]
        x_rope = x_rope * cos + _rotate_half(x_rope) * sin
        return torch.cat([x_rope, x_pass], dim=-1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        bsz, seq_len, dim = x.shape
        qkv = self.qkv(x)
        qkv = qkv.view(bsz, seq_len, 3, self.n_heads, self.head_dim)
        qkv = qkv.permute(0, 3, 1, 4, 2)
        q, k, v = qkv.unbind(dim=-1)

        if self.rope_dim > 0 and self.rope is not None:
            cos, sin = self.rope(seq_len, q.device, q.dtype)
            cos = cos[None, None, :, :]
            sin = sin[None, None, :, :]
            q = self._apply_rope(q, cos, sin)
            k = self._apply_rope(k, cos, sin)

        sdp_ctx = _sdp_kernel_context(self.sdp_kernel, q.device)
        input_dtype = q.dtype
        if q.device.type == "cuda" and q.dtype == torch.float32:
            q, k, v = q.bfloat16(), k.bfloat16(), v.bfloat16()
        with sdp_ctx:
            attn = F.scaled_dot_product_attention(
                q,
                k,
                v,
                dropout_p=self.dropout if self.training else 0.0,
                is_causal=True,
            )
        out = attn.to(input_dtype).permute(0, 2, 1, 3).reshape(bsz, seq_len, dim)
        out = self.proj(out)
        if self.dropout:
            out = F.dropout(out, p=self.dropout, training=self.training)
        return out


class MLP(nn.Module):
    def __init__(self, dim: int, ffn_dim: int, dropout: float) -> None:
        super().__init__()
        self.fc1 = nn.Linear(dim, ffn_dim, bias=False)
        self.fc2 = nn.Linear(ffn_dim, dim, bias=False)
        self.dropout = dropout

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = F.gelu(x)
        if self.dropout:
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.fc2(x)
        return x


class SwiGLUMLP(nn.Module):
    def __init__(self, dim: int, ffn_dim: int, dropout: float) -> None:
        super().__init__()
        self.gate = nn.Linear(dim, ffn_dim, bias=False)
        self.up = nn.Linear(dim, ffn_dim, bias=False)
        self.down = nn.Linear(ffn_dim, dim, bias=False)
        self.dropout = dropout

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        gate = F.silu(self.gate(x))
        up = self.up(x)
        x = gate * up
        if self.dropout:
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.down(x)
        return x


def sinkhorn_batched(inp: torch.Tensor, iters: int, eps: float) -> torch.Tensor:
    out = inp
    for _ in range(iters):
        out = out / (out.sum(dim=-1, keepdim=True) + eps)
        out = out / (out.sum(dim=-2, keepdim=True) + eps)
    return out


class MHCResidual(nn.Module):
    def __init__(
        self,
        hidden_dim: int,
        expansion_rate: int,
        sinkhorn_iters: int,
        alpha_init: float,
        rmsnorm_eps: float,
        sinkhorn_eps: float,
        layer_fn: nn.Module,
        use_dynamic_h: bool,
        use_fused_mhc: bool,
        recompute_mhc: bool,
    ) -> None:
        super().__init__()
        self.expansion_rate = expansion_rate
        self.sinkhorn_iters = sinkhorn_iters
        self.alpha_init = alpha_init
        self.rmsnorm_eps = rmsnorm_eps
        self.sinkhorn_eps = sinkhorn_eps
        self.layer_fn = layer_fn
        self.use_dynamic_h = use_dynamic_h
        self.use_fused_mhc = bool(use_fused_mhc and mhc_cuda is not None)
        self.recompute_mhc = recompute_mhc
        if self.use_fused_mhc and self.sinkhorn_eps != self.rmsnorm_eps:
            self.sinkhorn_eps = self.rmsnorm_eps
        self.rmsnorm = RMSNorm(hidden_dim, rmsnorm_eps)

        n = expansion_rate
        nC = n * hidden_dim

        if use_dynamic_h:
            self.phi_pre = nn.Parameter(torch.randn(n, nC) * 0.02)
            self.phi_post = nn.Parameter(torch.randn(n, nC) * 0.02)
            self.phi_res = nn.Parameter(torch.randn(n * n, nC) * 0.02)
            self.b_pre = nn.Parameter(torch.zeros(n))
            self.b_post = nn.Parameter(torch.zeros(n))
            self.b_res = nn.Parameter(torch.zeros(n, n))
            self.alpha_pre = nn.Parameter(torch.tensor(alpha_init))
            self.alpha_post = nn.Parameter(torch.tensor(alpha_init))
            self.alpha_res = nn.Parameter(torch.tensor(alpha_init))
        else:
            self.H_pre = nn.Parameter(torch.zeros(n))
            self.H_post = nn.Parameter(torch.zeros(n))
            self.H_res = nn.Parameter(torch.randn(n, n) * alpha_init)

    def _compute_dynamic_h_python(
        self, x_stream: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        bsz, seq_len, n, c = x_stream.shape
        nC = n * c
        x_flat = x_stream.reshape(bsz * seq_len, nC)
        rms = x_flat.pow(2).mean(dim=-1, keepdim=True).add(self.rmsnorm_eps).sqrt()
        x_norm = x_flat / rms

        phi_pre = self.phi_pre.float()
        phi_post = self.phi_post.float()
        phi_res = self.phi_res.float()

        p_pre = x_norm @ phi_pre.t()
        p_post = x_norm @ phi_post.t()
        p_res = x_norm @ phi_res.t()

        tilde_pre = self.alpha_pre.float() * p_pre + self.b_pre.float()
        tilde_post = self.alpha_post.float() * p_post + self.b_post.float()
        tilde_res = self.alpha_res.float() * p_res + self.b_res.float().view(1, n * n)
        tilde_res = torch.clamp(tilde_res, max=H_RES_EXP_CLAMP)

        H_pre = torch.sigmoid(tilde_pre)
        H_post = 2.0 * torch.sigmoid(tilde_post)

        H_res = tilde_res.view(bsz * seq_len, n, n)
        M = sinkhorn_batched(torch.exp(H_res), self.sinkhorn_iters, self.sinkhorn_eps)

        return H_pre, H_post, M

    def _compute_dynamic_h_fused(
        self, x_tokens: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        if not x_tokens.is_cuda:
            raise RuntimeError("fused mHC requires CUDA tensors")
        if mhc_cuda is None:
            raise RuntimeError("mhc_cuda is required for fused mHC")
        return DynamicHFusedFunction.apply(
            x_tokens,
            self.phi_pre,
            self.phi_post,
            self.phi_res,
            self.alpha_pre,
            self.alpha_post,
            self.alpha_res,
            self.b_pre,
            self.b_post,
            self.b_res,
            self.sinkhorn_iters,
            self.sinkhorn_eps,
        )

    def _compute_static_h(
        self, bsz: int, seq_len: int
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        n = self.expansion_rate
        H_pre = torch.sigmoid(self.H_pre.float()).view(1, n).expand(bsz * seq_len, n)
        H_post = 2.0 * torch.sigmoid(self.H_post.float()).view(1, n).expand(
            bsz * seq_len, n
        )
        H_res = torch.clamp(self.H_res.float(), max=H_RES_EXP_CLAMP)
        M = sinkhorn_batched(
            torch.exp(H_res).view(1, n, n).expand(bsz * seq_len, n, n),
            self.sinkhorn_iters,
            self.sinkhorn_eps,
        )
        return H_pre, H_post, M

    def _forward_recompute(self, x_stream: torch.Tensor) -> torch.Tensor:
        bsz, seq_len, n, c = x_stream.shape

        def _pre(x_in: torch.Tensor):
            x_f32 = x_in.float()
            if self.use_dynamic_h:
                H_pre, H_post, M = self._compute_dynamic_h_python(x_f32)
            else:
                H_pre, H_post, M = self._compute_static_h(bsz, seq_len)
            x_flat = x_f32.reshape(bsz * seq_len, n, c)
            x_agg = torch.einsum("bn,bnc->bc", H_pre, x_flat)
            x_agg = x_agg.view(bsz, seq_len, c)
            x_norm = self.rmsnorm(x_agg)
            return x_norm, H_post, M

        if self.training and torch.is_grad_enabled():
            x_norm, H_post, M = _checkpoint_fn(_pre, x_stream)
        else:
            x_norm, H_post, M = _pre(x_stream)

        layer_in = x_norm
        if layer_in.dtype != x_stream.dtype:
            layer_in = layer_in.to(x_stream.dtype)
        y = self.layer_fn(layer_in)

        def _mix(
            x_in: torch.Tensor,
            y_in: torch.Tensor,
            H_post_in: torch.Tensor,
            M_in: torch.Tensor,
        ) -> torch.Tensor:
            x_f32 = x_in.float()
            x_flat = x_f32.reshape(bsz * seq_len, n, c)
            y_flat = y_in.reshape(bsz * seq_len, c).float()
            y_dist = H_post_in.unsqueeze(-1) * y_flat.unsqueeze(1)
            x_mixed = torch.einsum("bij,bjc->bic", M_in, x_flat)
            out = x_mixed + y_dist
            return out.view(bsz, seq_len, n, c)

        if self.training and torch.is_grad_enabled():
            out = _checkpoint_fn(_mix, x_stream, y, H_post, M)
        else:
            out = _mix(x_stream, y, H_post, M)
        return out.to(x_stream.dtype)

    def forward(self, x_stream: torch.Tensor) -> torch.Tensor:
        if self.recompute_mhc and self.training and torch.is_grad_enabled():
            return self._forward_recompute(x_stream)
        bsz, seq_len, n, c = x_stream.shape
        use_fused = self.use_fused_mhc and x_stream.is_cuda and mhc_cuda is not None
        if use_fused:
            x_tokens = x_stream.reshape(bsz * seq_len, n, c).contiguous()
        else:
            x_stream_f32 = x_stream.float()
            x_tokens = x_stream_f32.reshape(bsz * seq_len, n, c).contiguous()
        if self.use_dynamic_h:
            if use_fused:
                H_pre, H_post, M = self._compute_dynamic_h_fused(x_tokens)
            else:
                H_pre, H_post, M = self._compute_dynamic_h_python(x_stream_f32)
        else:
            H_pre, H_post, M = self._compute_static_h(bsz, seq_len)

        if use_fused:
            H_pre = H_pre.contiguous()
            H_post = H_post.contiguous()
            M = M.contiguous()
            x_agg_flat = AggregateFusedFunction.apply(x_tokens, H_pre)
            if mhc_ops is not None:
                x_norm_flat = mhc_ops.rmsnorm(
                    x_agg_flat, self.rmsnorm.weight, self.rmsnorm_eps
                )
            else:
                x_norm_flat = self.rmsnorm(x_agg_flat.float())
            layer_in = x_norm_flat.reshape(bsz, seq_len, c)
            if layer_in.dtype != x_stream.dtype:
                layer_in = layer_in.to(x_stream.dtype)
            y = self.layer_fn(layer_in)
            y_flat = y.reshape(bsz * seq_len, c).contiguous()
            y_bf16 = y_flat.to(torch.bfloat16).contiguous()
            out_flat = MixFusedFunction.apply(x_tokens, y_bf16, H_post, M)
            return out_flat.reshape(bsz, seq_len, n, c).to(x_stream.dtype)

        x_flat = x_tokens
        x_agg = torch.einsum("bn,bnc->bc", H_pre, x_flat)
        x_agg = x_agg.view(bsz, seq_len, c)
        x_norm = self.rmsnorm(x_agg)
        layer_in = x_norm
        if layer_in.dtype != x_stream.dtype:
            layer_in = layer_in.to(x_stream.dtype)
        y = self.layer_fn(layer_in)
        y_flat = y.reshape(bsz * seq_len, c).float()

        y_dist = H_post.unsqueeze(-1) * y_flat.unsqueeze(1)
        x_mixed = torch.einsum("bij,bjc->bic", M, x_flat)
        out = x_mixed + y_dist
        return out.view(bsz, seq_len, n, c).to(x_stream.dtype)


class MHCTransformerBlock(nn.Module):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__()
        attn = CausalSelfAttention(
            config.hidden_dim,
            config.n_heads,
            config.rope_dim,
            config.rope_theta,
            config.dropout,
            config.sdp_kernel,
        )
        if config.mlp_type == "swiglu":
            mlp = SwiGLUMLP(config.hidden_dim, config.ffn_dim, config.dropout)
        else:
            mlp = MLP(config.hidden_dim, config.ffn_dim, config.dropout)

        self.attn_resid = MHCResidual(
            config.hidden_dim,
            config.expansion_rate,
            config.sinkhorn_iters,
            config.alpha_init,
            config.rmsnorm_eps,
            config.sinkhorn_eps,
            attn,
            config.use_dynamic_h,
            config.use_fused_mhc,
            config.recompute_mhc,
        )
        self.mlp_resid = MHCResidual(
            config.hidden_dim,
            config.expansion_rate,
            config.sinkhorn_iters,
            config.alpha_init,
            config.rmsnorm_eps,
            config.sinkhorn_eps,
            mlp,
            config.use_dynamic_h,
            config.use_fused_mhc,
            config.recompute_mhc,
        )

    def forward(self, x_stream: torch.Tensor) -> torch.Tensor:
        x_stream = self.attn_resid(x_stream)
        x_stream = self.mlp_resid(x_stream)
        return x_stream


class MHCTransformer(nn.Module):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__()
        self.config = config
        self.recompute_ratio = max(0.0, min(1.0, config.recompute_ratio))
        self.token_emb = nn.Embedding(config.vocab_size, config.hidden_dim)
        self.blocks = nn.ModuleList(
            [MHCTransformerBlock(config) for _ in range(config.n_layers)]
        )
        self.final_norm = RMSNorm(config.hidden_dim, config.rmsnorm_eps)
        self.lm_head = nn.Linear(config.hidden_dim, config.vocab_size, bias=False)
        self.lm_head.weight = self.token_emb.weight
        self._checkpoint_indices = self._build_checkpoint_indices(
            len(self.blocks), self.recompute_ratio
        )
        self._init_weights()

    def _init_weights(self) -> None:
        """Initialize weights following DeepSeek/LLaMA conventions."""
        init_std = 0.02
        depth_std = init_std / (2 * self.config.n_layers) ** 0.5

        for module in self.modules():
            if module is self:
                continue
            if isinstance(module, nn.Embedding):
                nn.init.normal_(module.weight, mean=0.0, std=init_std)
            elif isinstance(module, nn.Linear):
                nn.init.normal_(module.weight, mean=0.0, std=init_std)
                if module.bias is not None:
                    nn.init.zeros_(module.bias)

        # Scale residual-path projections by 1/sqrt(2*n_layers) to
        # prevent signal growth through the residual stream.
        for block in self.blocks:
            for resid in (block.attn_resid, block.mlp_resid):
                layer_fn = resid.layer_fn
                if isinstance(layer_fn, CausalSelfAttention):
                    nn.init.normal_(layer_fn.proj.weight, mean=0.0, std=depth_std)
                elif isinstance(layer_fn, SwiGLUMLP):
                    nn.init.normal_(layer_fn.down.weight, mean=0.0, std=depth_std)
                elif isinstance(layer_fn, MLP):
                    nn.init.normal_(layer_fn.fc2.weight, mean=0.0, std=depth_std)

    @staticmethod
    def _build_checkpoint_indices(num_layers: int, ratio: float) -> set[int]:
        if ratio <= 0.0 or num_layers <= 0:
            return set()
        if ratio >= 1.0:
            return set(range(num_layers))
        num_checkpoint = max(1, int(round(num_layers * ratio)))
        indices = set()
        for i in range(num_checkpoint):
            idx = int(i * num_layers / num_checkpoint)
            indices.add(min(idx, num_layers - 1))
        return indices

    def forward(
        self, input_ids: torch.Tensor, return_hidden: bool = False
    ) -> torch.Tensor:
        bsz, seq_len = input_ids.shape
        if seq_len > self.config.max_seq_len:
            raise ValueError("seq_len exceeds config.max_seq_len")
        x = self.token_emb(input_ids)
        x_stream = (
            x.unsqueeze(2)
            .expand(bsz, seq_len, self.config.expansion_rate, self.config.hidden_dim)
            .contiguous()
        )
        use_checkpoint = (
            self.recompute_ratio > 0.0
            and self.training
            and torch.is_grad_enabled()
            and self._checkpoint_indices
        )
        if use_checkpoint:
            for idx, block in enumerate(self.blocks):
                if idx in self._checkpoint_indices:
                    x_stream = _checkpoint_block(block, x_stream)
                else:
                    x_stream = block(x_stream)
        else:
            for block in self.blocks:
                x_stream = block(x_stream)
        x_out = x_stream.mean(dim=2)
        x_out = self.final_norm(x_out)
        if return_hidden:
            return x_out
        return self.lm_head(x_out)



__version__ = "0.1.0"

from .layer import MHCLayer

try:
    from .ops import (
        sinkhorn_knopp,
        rmsnorm,
        mhc_layer_fused,
        mhc_layer_fused_dynamic,
        mhc_layer_fused_dynamic_inference,
    )
except ImportError:  # mhc_cuda not built; fused ops unavailable
    sinkhorn_knopp = None
    rmsnorm = None
    mhc_layer_fused = None
    mhc_layer_fused_dynamic = None
    mhc_layer_fused_dynamic_inference = None

__all__ = [
    "__version__",
    "MHCLayer",
    "sinkhorn_knopp",
    "rmsnorm",
    "mhc_layer_fused",
    "mhc_layer_fused_dynamic",
    "mhc_layer_fused_dynamic_inference",
]



import os
import sys
from setuptools import setup, find_packages


def _is_sdist_build():
    """Check if we are building a source distribution (no compilation needed)."""
    return "sdist" in sys.argv or "egg_info" in sys.argv


def get_cuda_arch_flags():
    import torch

    cuda_arch = os.environ.get("CUDA_ARCH")
    if cuda_arch:
        return [f"-gencode=arch=compute_{cuda_arch},code=sm_{cuda_arch}"]
    if not torch.cuda.is_available():
        raise RuntimeError(
            "CUDA is required to build mhc (set CUDA_ARCH env var for headless builds)"
        )
    major, minor = torch.cuda.get_device_capability()
    return [f"-gencode=arch=compute_{major}{minor},code=sm_{major}{minor}"]


def get_extra_defines():
    import torch

    cuda_arch = os.environ.get("CUDA_ARCH")
    if cuda_arch:
        defines = []
        if int(cuda_arch) >= 90:
            defines.append("-DMHC_ENABLE_PDL")
        return defines
    if not torch.cuda.is_available():
        return []
    major, _ = torch.cuda.get_device_capability()
    defines = []
    if major >= 9:
        defines.append("-DMHC_ENABLE_PDL")
    return defines


# Skip CUDA extension entirely for sdist builds (no CUDA toolkit needed)
if _is_sdist_build():
    ext_modules = []
    cmdclass = {}
else:
    from torch.utils.cpp_extension import BuildExtension, CUDAExtension

    ext_modules = [
        CUDAExtension(
            name="mhc_cuda",
            sources=["src/python/bindings.cu"],
            include_dirs=[
                os.path.join(
                    os.path.dirname(os.path.abspath(__file__)), "src/csrc/include"
                ),
                os.path.join(
                    os.path.dirname(os.path.abspath(__file__)), "src/csrc/kernels"
                ),
            ],
            extra_compile_args={
                "cxx": ["-O3"],
                "nvcc": [
                    "-O3",
                    "--expt-relaxed-constexpr",
                    "-U__CUDA_NO_HALF_OPERATORS__",
                    "-U__CUDA_NO_HALF_CONVERSIONS__",
                    "-U__CUDA_NO_BFLOAT16_CONVERSIONS__",
                    "--use_fast_math",
                ]
                + get_cuda_arch_flags()
                + get_extra_defines(),
            },
            libraries=["cublas", "cublasLt"],
        )
    ]
    cmdclass = {"build_ext": BuildExtension}

setup(
    name="mhc-cuda",
    version="0.1.0",
    description="CUDA implementation of Manifold-Constrained Hyper-Connections",
    long_description=open("README.md").read() if os.path.exists("README.md") else "",
    long_description_content_type="text/markdown",
    author="Andre Slavescu",
    url="https://github.com/AndreSlavescu/mHC.cu",
    packages=find_packages(where="src/python"),
    package_dir={"": "src/python"},
    ext_modules=ext_modules,
    cmdclass=cmdclass,
    python_requires=">=3.10",
    install_requires=[
        "accelerate>=0.26.0",
        "datasets",
        "safetensors",
        "sentencepiece",
        "tiktoken",
        "torch>=2.0.0",
        "tqdm",
        "transformers",
    ],
    extras_require={
        "dev": ["black", "pytest", "ruff"],
    },
)
